`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date: 09/23/2023 03:41:41 PM
// Design Name: 
// Module Name: top
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module top(
    input clock,            // system clock
    input reset,            // BTNR
    input version,          // BTNU
    input adc_n, adc_p,     // VCAUX 6, P and N
    input [15:0] sw,        // slide switches
    output [15:0] led,      // 16  onboard LEDs above the switches
    output [6:0] segment,   // 4 digit LED display
    output dp,              // "." next to each LED digit
    output [3:0] digit,     // which of the 4 digits to drive
    output [7:0] JB,        // PMOD for debugging
    input rx,               // UART receive
    output tx               // UART transmit
    );
    parameter VERSION = 'h0001;
    //
    //  make a slower clock for the UART transmitter and for the FSM
    //  below that controls it.   We do this because the logic analyzer
    //  is slow too, it will only look at 8 wires at 25MHz.   So if we
    //  divide the system clock by 4, that will give us a 25MHz clock
    //  and that should do nicelly, and it will make clocks_per_bit come
    //  out as an integer
    //
    reg [1:0] count2 = 0;
    always @ (posedge clock) count2 <= count2 + 1;
    wire clock25;
    BUFG clk25buf (.I(count2[1]), .O(clock25) );
    //
    //  generate a 104MHz clock for the XADC so we can get samples at 1MHz
    //
    wire locked, clock104;
    clock_104 XADC_CLOCK(
        .reset(reset),
        .locked(locked),
        .clk_in1(clock),
        .clk_out1(clock104)
    );
    //
    //  next drive the 4 7-segment displays
    //
    wire [15:0] display_this;
    display4 DISPLAY (
        .clk100(clock),
        .number(display_this),
        .digit(digit),
        .segments(segment),
        .period(dp)
        );
    //
    //  instantiate the UART receiver.  run with the 50MHz clock so that
    //  we can run the transmitter FSM at 100MHz and not get lost
    //
    wire dv;
    wire [7:0] rx_data;
    uart_rx RECEIVER (
        .i_Clocks_per_Bit('d25),
        .i_Clock(clock25),
        .i_Reset(reset),
        .i_Rx_Serial(rx),       // tied to FPGA rx output
        .o_Rx_DV(dv),
        .o_Rx_Byte(rx_data)
        );
    //
    //  instantiate the UART transmitter
    //
    wire tx_active, tx_done;
    wire do_transmit;
    wire [7:0] transmit_byte;
    uart_tx TRANSMITTER (
        .i_Clocks_per_Bit('d25),
        .i_Clock(clock25),
        .i_Reset(reset),
        .i_Tx_DV(do_transmit),
        .i_Tx_Byte(transmit_byte), 
        .o_Tx_Active(tx_active),
        .o_Tx_Serial(tx),       // tied to FPGA tx output
        .o_Tx_Done(tx_done)
        );
    //
    //  here is the XADC block
    //
    wire [6:0] daddr_in = 7'h16;
    wire adc_ready, isbusy, adc_data_ready, eos_out, alarm;
    wire [15:0] adc_data;
    wire [4:0] channel_out;
    myxadc XADC_INST (
        .daddr_in(7'h16),   // specifies vcaux6 pints to digitize
        .dclk_in(clock104),    // 50MHz clock
        .den_in(adc_ready), // tied to adc_ready, tells adc to convert, tieing causes continuous conversions
        .di_in(16'h0),      // to set the data to something, not used here
        .dwe_in(1'b0),      //  set to enable writing to di_in, which we don't want to do
        .vauxp6(adc_p),     //  positive input to digitize
        .vauxn6(adc_n),     //  negative input to digitize
        .busy_out(isbusy),  // tells you the adc is busy converting
        .channel_out(channel_out[4:0]), // for using more than 1 channel, tells you which one.  not used here
        .do_out(adc_data),      // adc value from conversion
        .drdy_out(adc_data_ready),  //tells you valid data is ready to be latched
        .eoc_out(adc_ready),   //  specifies that the ADC is ready (conversion complete)
        .eos_out(eos_out),     //  specifies that conversion sequence is complete
        .alarm_out(alarm),      // OR's output of all internal alarms, not used here
        .vp_in(1'b0),           // dedicated analog input pair for differential, tied to 0 if not used
        .vn_in(1'b0)
    );
    //
    // wait for XADC to tell you something is ready to latch. note this means continuous latching
    //
    reg [15:0] r_adc_data;
    always @ (negedge isbusy) begin
        if (reset) r_adc_data <= 16'h0;
        else r_adc_data <= adc_data;
        end
    //
    //  make a ~1Hz clock so we can run the LED display slower
    //
    reg [26:0] counter;
    reg [15:0] s_adc_data;
    always @ (posedge clock) begin
        if (reset) counter <= 0;
        else counter <= counter + 1;
        end
    wire clock_1hz = counter[26];
    always @ (posedge clock_1hz) s_adc_data <= r_adc_data;
    assign display_this = version ? VERSION : s_adc_data;
    //
    //  now make a state machine to deal with transmitting 2 bytes
    //
    reg [2:0] tx_state;     // 7 states so 3 bits will do
    localparam [2:0] TX_WAIT=0, TX_BYTE1=1, TX_DO1=2, TX_WAIT1=3, TX_BYTE2=4,
                    TX_DO2=5, TX_WAIT2=6;
    reg doit;
    wire [16:0] transmit_word;
    wire begin_transfer;
    reg [7:0] tx_data;
    always @ (posedge clock25) begin
        if (reset) begin
            tx_state <= TX_WAIT;
            doit <= 0;
            tx_data <= 0;
        end
        else 
            case (tx_state)
                TX_WAIT: begin
                    if (begin_transfer) tx_state <= TX_BYTE1;
                    else tx_state <= TX_WAIT;
                    doit <= 0;
                    tx_data <= 0;
                end
                TX_BYTE1: begin
                    tx_data <= transmit_word[7:0];
                    tx_state <= TX_DO1;
                end
                TX_DO1: begin
                    doit <= 1;
                    tx_state <= TX_WAIT1;
                end
                TX_WAIT1: begin
                    doit <= 0;
                    if (tx_done) tx_state <= TX_BYTE2;
                    else tx_state <= TX_WAIT1;
                end
                TX_BYTE2: begin
                    tx_data <= transmit_word[15:8];
                    tx_state <= TX_DO2;
                end
                TX_DO2: begin
                    doit <= 1;
                    tx_state <= TX_WAIT2;
                end
                TX_WAIT2: begin
                    doit <= 0;
                    if (tx_done) tx_state <= TX_WAIT;
                    else tx_state <= TX_WAIT2;
                end
            endcase
    end
    assign do_transmit = doit;
    assign transmit_byte = tx_data;
    //
    //  next comes the uart receive state machine
    //
    reg [2:0] rx_state;
    localparam [2:0] RX_WAIT=0, RX_RPI_LATCH=1, RX_Transfer=2,
       RX_PAUSE=3, RX_WAIT_Done1=4, RX_WAIT_Done2=5;
    reg start_tx;
    wire [1:0] latch_what = {rx_data[2],rx_data[1]};
    reg [15:0] transmit_this;
    always @ (posedge clock25) begin
        if (reset) begin
            start_tx <= 0;
            rx_state <= RX_WAIT;
            transmit_this <= 0;
        end
        else 
            case (rx_state)
                RX_WAIT: begin
                    start_tx <= 0;
                    transmit_this <= 0;
                    if (dv && rx_data[0]) rx_state <= RX_RPI_LATCH;
                    else rx_state <= RX_WAIT;
                end
                RX_RPI_LATCH: begin
                    rx_state <= RX_Transfer;
                    case (latch_what) 
                        2'b00: transmit_this <= 'hEEEE;          // error!
                        2'b01: transmit_this <= r_adc_data;
                        2'b10: transmit_this <= sw;
                        2'b11: transmit_this <= VERSION;
                    endcase
                end
                RX_Transfer: begin
                    start_tx <= 1;
                    rx_state <= RX_WAIT_Done1;                    
                end
                RX_WAIT_Done1: begin
                    if (tx_done) rx_state <= RX_PAUSE;
                    else rx_state <= RX_WAIT_Done1;
                end
                RX_PAUSE: begin
                    rx_state <= RX_WAIT_Done2;
                end
                RX_WAIT_Done2: begin
                    if (tx_done) rx_state <= RX_WAIT_Done2;
                    else rx_state <= RX_WAIT;
                end
            endcase
    end
    assign transmit_word = transmit_this;
    assign begin_transfer = start_tx;
    assign led = transmit_this;
    assign JB = {doit,do_transmit,tx_active,tx_done,latch_what[1:0],dv,clock25};
endmodule
